/*
     RUN_ELF - Load and execute big endian ix86 ELF binaries
     Copyright (C) 2001-2002 Martin Blom <martin@blom.org>
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <string.h>

#define bswap_long(x) \
     ((((x) & 0xff000000) >> 24) | (((x) & 0x00ff0000) >>  8) | \
      (((x) & 0x0000ff00) <<  8) | (((x) & 0x000000ff) << 24))

#define bswap_short(x) \
     ((((x) >> 8) & 0xff) | (((x) & 0xff) << 8))

#ifndef __amiga__
# include <unistd.h>
# include <fcntl.h>
# include <stdlib.h>
# include <stdio.h>

# define AllocVec(size,type) malloc(size)
# define FreeVec(addr)       free(addr)
# define MEMF_PUBLIC         0

# define Open(name,mode)     open(name,mode)
# define Close(file)         close(file)
# define Read(file,buf,len)  read(file,buf,len)
# define Seek(file,pos,mode) lseek(file,pos,mode)
# define MODE_OLDFILE        O_RDONLY
# define OFFSET_BEGINNING    SEEK_SET
# define BPTR_NULL           -1UL

# define big_16(x) bswap_short(x)
# define big_32(x) bswap_long(x)

typedef unsigned long BPTR;

#else

# include <exec/memory.h>
# include <proto/exec.h>

# define big_16(x) (x)
# define big_32(x) (x)

#endif


#include "new_elfloader.h"
#include <elf.h>

#ifndef __amiga__
int
main( int argc, char** argv )
{
  int error;
  
  if( argc != 2 )
  {
    fprintf( stderr, "Usage: %s file\n", argv[ 0 ] );
    return 1;
  }
  
  ELFLoadSeg( argv[ 1 ], &error );

  printf( "rc: %d\n", error );
  
  return error;
}

struct DosSegmentDescr
{
    unsigned long base;
    unsigned long size;
};

//void*
//AllocDosSegments( unsigned long alignment,

#endif


BPTR
ELFLoadSeg( const char* file, int* error_code )
{
  BPTR       res = 0;
  BPTR       f;
  Elf32_Ehdr header;

  *error_code = NEWELF_OK;
  
  f = Open( file, MODE_OLDFILE );

  if( f == BPTR_NULL )
  {
    *error_code = NEWELF_OPEN_FAILED;
  }
  else
  {
    if( Read( f, &header, sizeof( header ) ) != sizeof( header ) )
    {
      *error_code = NEWELF_INVALID_FILE;
    }
    else
    {
      if( strncmp( header.e_ident, ELFMAG, SELFMAG ) != 0 ||
	  header.e_ident[ EI_CLASS ] != ELFCLASS32 ||
	  header.e_ident[ EI_DATA ] != ELFDATA2MSB ||
	  header.e_ident[ EI_VERSION ] != EV_CURRENT ||
	  header.e_ident[ EI_OSABI ] != ELFOSABI_NONE ||
//	  header.e_ident[ EI_OSABI ] != ELFOSABI_STANDALONE ||
	  big_16( header.e_type ) != ET_EXEC ||
	  big_16( header.e_machine ) != EM_386 ||
	  big_32( header.e_version ) != EV_CURRENT )
      {
	*error_code = NEWELF_INVALID_TYPE;
      }
      else
      {
	Elf32_Phdr* prog_headers;
	size_t      prog_headers_size = ( big_16( header.e_phentsize) *
					  big_16( header.e_phnum ) );

	Seek( f, big_32( header.e_phoff ), OFFSET_BEGINNING );

	prog_headers = AllocVec( prog_headers_size, MEMF_PUBLIC );

	if( prog_headers == NULL )
	{
	  *error_code = NEWELF_NO_MEMORY;
	}
	else
	{
	  if( Read( f, prog_headers, prog_headers_size ) != prog_headers_size )
	  {
	    *error_code = NEWELF_INVALID_FILE;
	  }
	  else
	  {
	    int            i;
	    size_t         ph_size = big_16( header.e_phentsize);
	    Elf32_Phdr*    ph;
	    unsigned long  dynamic_vaddr = 0;
	    unsigned long  interpr_vaddr = 0;
	    unsigned long  alignment     = 0;
	    unsigned long  segment_vaddr = ~0UL;
	    unsigned long  segment_end   = 0;
	    
	    for( i = 0, ph = prog_headers;
		 i < big_16( header.e_phnum );
		 ++i, ph = (Elf32_Phdr*) ((char*) ph + ph_size) )
	    {
	      // We assume power-of-two alignment
	      if( big_32( ph->p_align ) > alignment )
	      {
		alignment = big_32( ph->p_align );
	      }
	      
	      switch( big_32( ph->p_type ) )
	      {
		case PT_PHDR:
		  // We've already got it
		  break;

		case PT_LOAD:
		  if( big_32( ph->p_vaddr ) < segment_vaddr )
		  {
		    segment_vaddr = big_32( ph->p_vaddr );
		  }

		  if( big_32( ph->p_vaddr ) + big_32( ph->p_memsz ) >
		      segment_end )
		  {
		    segment_end = ( big_32( ph->p_vaddr ) +
				    big_32( ph->p_memsz ) );
		  }
		  
		  break;

		case PT_DYNAMIC:
		  dynamic_vaddr = big_32( ph->p_vaddr );
		  break;

		case PT_INTERP:
		  interpr_vaddr = big_32( ph->p_vaddr );
		  break;
		  
		default:
		  printf( "type: %d\n", big_32( ph->p_type ) );
		  break;
	      }
	    }

	    printf("start: %x len %x align: %x dyn: %x ip: %x\n",
		   segment_vaddr, segment_end-segment_vaddr, alignment,
		   dynamic_vaddr, interpr_vaddr );
	  }
	  
	  FreeVec( prog_headers );
	}
      }
    }
    
    Close( f );
  }

  return res;
}
